@@ -20,6 +20,11 @@
 import mutagen.apev2
 import mutagen.mp3
 import mutagen.trueaudio
+try:
+    import mutagen.aiff
+except ImportError:
+    mutagen.aiff = None
+
 import re
 from collections import defaultdict
 from mutagen import id3
@@ -32,39 +37,6 @@ from picard.util import encode_filename, sanitize_date
 from urlparse import urlparse
 
 
-# Ugly, but... I need to save the text in ISO-8859-1 even if it contains
-# unsupported characters and this better than encoding, decoding and
-# again encoding.
-def patched_EncodedTextSpec_write(self, frame, value):
-    try:
-        enc, term = self._encodings[frame.encoding]
-    except AttributeError:
-        enc, term = self.encodings[frame.encoding]
-    return value.encode(enc, 'ignore') + term
-
-id3.EncodedTextSpec.write = patched_EncodedTextSpec_write
-
-
-# One more "monkey patch". The ID3 spec says that multiple text
-# values should be _separated_ by the string terminator, which
-# means that e.g. 'a\x00' are two values, 'a' and ''.
-def patched_MultiSpec_write(self, frame, value):
-    data = self._write_orig(frame, value)
-    spec = self.specs[-1]
-    if isinstance(spec, id3.EncodedTextSpec):
-        try:
-            term = spec._encodings[frame.encoding][1]
-        except AttributeError:
-            term = spec.encodings[frame.encoding][1]
-        if data.endswith(term):
-            data = data[:-len(term)]
-    return data
-
-
-id3.MultiSpec._write_orig = id3.MultiSpec.write
-id3.MultiSpec.write = patched_MultiSpec_write
-
-
 id3.TCMP = compatid3.TCMP
 id3.TSO2 = compatid3.TSO2
 id3.TSOC = compatid3.TSOC
@@ -85,6 +57,16 @@ __ID3_IMAGE_TYPE_MAP = {
 __ID3_REVERSE_IMAGE_TYPE_MAP = dict([(v, k) for k, v in __ID3_IMAGE_TYPE_MAP.iteritems()])
 
 
+def id3text(text, encoding):
+    """Returns a string which only contains code points which can
+    be encododed with the given numeric id3 encoding.
+    """
+
+    if encoding == 0:
+        return text.encode("latin1", "replace").decode("latin1")
+    return text
+
+
 def image_type_from_id3_num(id3type):
     return __ID3_REVERSE_IMAGE_TYPE_MAP.get(id3type, "other")
 
@@ -100,13 +82,13 @@ def types_from_id3(id3type):
 class ID3File(File):
 
     """Generic ID3-based file."""
-    _File = None
     _IsMP3 = False
 
     __upgrade = {
         'XSOP': 'TSOP',
         'TXXX:ALBUMARTISTSORT': 'TSO2',
         'TXXX:COMPOSERSORT': 'TSOC',
+        'TXXX:mood': 'TMOO',
     }
 
     __translate = {
@@ -125,6 +107,7 @@ class ID3File(File):
         'TCOM': 'composer',
         'TENC': 'encodedby',
         'TBPM': 'bpm',
+        'TKEY': 'key',
         'TLAN': 'language',
         'TCON': 'genre',
         'TMED': 'media',
@@ -194,7 +177,7 @@ class ID3File(File):
 
     def _load(self, filename):
         log.debug("Loading file %r", filename)
-        file = self._File(encode_filename(filename), ID3=compatid3.CompatID3)
+        file = self._get_file(encode_filename(filename))
         tags = file.tags or {}
         # upgrade custom 2.3 frames to 2.4
         for old, new in self.__upgrade.items():
@@ -290,20 +273,13 @@ class ID3File(File):
     def _save(self, filename, metadata):
         """Save metadata to the file."""
         log.debug("Saving file %r", filename)
-        try:
-            tags = compatid3.CompatID3(encode_filename(filename))
-        except mutagen.id3.ID3NoHeaderError:
-            tags = compatid3.CompatID3()
+        tags = self._get_tags(filename)
 
         if config.setting['clear_existing_tags']:
             tags.clear()
         if metadata.images_to_be_saved_to_tags:
             tags.delall('APIC')
 
-        if config.setting['write_id3v1']:
-            v1 = 2
-        else:
-            v1 = 0
         encoding = {'utf-8': 3, 'utf-16': 1}.get(config.setting['id3v2_encoding'], 0)
 
         if 'tracknumber' in metadata:
@@ -311,14 +287,14 @@ class ID3File(File):
                 text = '%s/%s' % (metadata['tracknumber'], metadata['totaltracks'])
             else:
                 text = metadata['tracknumber']
-            tags.add(id3.TRCK(encoding=0, text=text))
+            tags.add(id3.TRCK(encoding=0, text=id3text(text, 0)))
 
         if 'discnumber' in metadata:
             if 'totaldiscs' in metadata:
                 text = '%s/%s' % (metadata['discnumber'], metadata['totaldiscs'])
             else:
                 text = metadata['discnumber']
-            tags.add(id3.TPOS(encoding=0, text=text))
+            tags.add(id3.TPOS(encoding=0, text=id3text(text, 0)))
 
         # This is necessary because mutagens HashKey for APIC frames only
         # includes the FrameID (APIC) and description - it's basically
@@ -336,7 +312,7 @@ class ID3File(File):
             tags.add(id3.APIC(encoding=0,
                               mime=image.mimetype,
                               type=image_type_as_id3_num(image.maintype),
-                              desc=desctag,
+                              desc=id3text(desctag, 0),
                               data=image.data))
 
         tmcl = mutagen.id3.TMCL(encoding=encoding, people=[])
@@ -344,6 +320,9 @@ class ID3File(File):
 
         tags.delall('TCMP')
         for name, values in metadata.rawitems():
+            values = [id3text(v, encoding) for v in values]
+            name = id3text(name, encoding)
+
             if name.startswith('performer:'):
                 role = name.split(':', 1)[1]
                 for value in values:
@@ -393,6 +372,11 @@ class ID3File(File):
                         for url in values:
                             tags.add(id3.WOAR(url=url))
                 elif frameid.startswith('T'):
+                    if config.setting['write_id3v23']:
+                        if frameid == 'TMOO':
+                            tags.add(id3.TXXX(encoding=encoding, desc='mood', text=values))
+                    # No need to care about the TMOO tag being added again as it is
+                    # automatically deleted by Mutagen if id2v23 is selected
                     tags.add(getattr(id3, frameid)(encoding=encoding, text=values))
                     if frameid == 'TSOA':
                         tags.delall('XSOA')
@@ -411,20 +395,15 @@ class ID3File(File):
                     if frameclass:
                         tags.add(frameclass(encoding=encoding, text=values))
             # don't save private / already stored tags
-            elif not name.startswith("~") and not name in self.__other_supported_tags:
+            elif not name.startswith("~") and name not in self.__other_supported_tags:
                 tags.add(id3.TXXX(encoding=encoding, desc=name, text=values))
 
-        if tmcl.people:
-            tags.add(tmcl)
-        if tipl.people:
-            tags.add(tipl)
+        tags.add(tmcl)
+        tags.add(tipl)
 
-        if config.setting['write_id3v23']:
-            tags.update_to_v23(join_with=config.setting['id3v23_join_with'])
-            tags.save(encode_filename(filename), v2=3, v1=v1)
-        else:
-            tags.update_to_v24()
-            tags.save(encode_filename(filename), v2=4, v1=v1)
+        self._remove_deleted_tags(metadata, tags)
+
+        self._save_tags(tags, encode_filename(filename))
 
         if self._IsMP3 and config.setting["remove_ape_from_mp3"]:
             try:
@@ -432,11 +411,108 @@ class ID3File(File):
             except:
                 pass
 
+    def _remove_deleted_tags(self, metadata, tags):
+        """Remove the tags from the file that were deleted in the UI"""
+        for name in metadata.deleted_tags:
+            real_name = self._get_tag_name(name)
+            try:
+                if name.startswith('performer:'):
+                    role = name.split(':', 1)[1]
+                    for key, frame in tags.items():
+                        if frame.FrameID in ('TMCL', 'TIPL', 'IPLS'):
+                            for people in frame.people:
+                                if people[0] == role:
+                                    frame.people.remove(people)   
+                elif name.startswith('comment:'):
+                    desc = name.split(':', 1)[1]
+                    if desc.lower()[:4] != 'itun':
+                        for key, frame in tags.items():
+                            if frame.FrameID == 'COMM' and frame.desc == desc:
+                                del tags[key]
+                elif name.startswith('lyrics:') or name == 'lyrics':
+                    if ':' in name:
+                        desc = name.split(':', 1)[1]
+                    else:
+                        desc = ''
+                    for key, frame in tags.items():
+                        if frame.FrameID == desc:
+                            del tags[key]
+                elif name in self._rtipl_roles:
+                    role = self._rtipl_roles[name]
+                    for key, frame in tags.items():
+                        if frame.FrameID in ('TIPL', 'IPLS'):
+                            for people in frame.people:
+                                if people[0] == role:
+                                    frame.people.remove(people)   
+                elif name == 'musicbrainz_recordingid':
+                    for key, frame in tags.items():
+                        if frame.FrameID == 'UFID' and frame.owner == 'http://musicbrainz.org':
+                            del tags[key]
+                elif real_name == 'POPM':
+                    for key, frame in tags.items():
+                        if frame.FrameID == 'POPM' and frame.email == config.setting['rating_user_email']:
+                            del tags[key]
+                elif real_name in self.__translate:
+                    del tags[real_name]
+                elif real_name in self.__translate_freetext:
+                    for key, frame in tags.items():
+                        if frame.FrameID == 'TXXX' and frame.desc == real_name:
+                            del tags[key]
+                elif not name.startswith("~") and name not in self.__other_supported_tags:
+                    for key, frame in tags.items():
+                        if frame.FrameID == 'TXXX' and frame.desc == name:
+                            del tags[key]
+                elif name.startswith("~"):
+                    name = name[1:]
+                    for key, frame in tags.items():
+                        if frame.FrameID == 'TXXX' and frame.desc == name:
+                            del tags[key]
+                elif name in self.__other_supported_tags:
+                    del tags[real_name]
+            except KeyError:
+                pass
+
     def supports_tag(self, name):
-        return name in self.__rtranslate or name in self.__rtranslate_freetext\
-            or name.startswith('performer:')\
-            or name.startswith('lyrics:') or name == 'lyrics'\
-            or name in self.__other_supported_tags
+        return ((name and not name.startswith("~"))
+                or name in ("~rating", "~length")
+                or name.startswith("~id3"))
+
+    def _get_tag_name(self, name):
+        if name in self.__rtranslate:
+            return self.__rtranslate[name]
+        elif name in self.__rtranslate_freetext:
+            return self.__rtranslate_freetext[name]
+        elif name == '~rating':
+            return 'POPM'
+        elif name == 'tracknumber':
+            return 'TRCK'
+        elif name == 'discnumber':
+            return 'TPOS'
+        else:
+            return None
+
+    def _get_file(self, filename):
+        raise NotImplementedError()
+
+    def _get_tags(self, filename):
+        try:
+            return compatid3.CompatID3(encode_filename(filename))
+        except mutagen.id3.ID3NoHeaderError:
+            return compatid3.CompatID3()
+
+    def _save_tags(self, tags, filename):
+        if config.setting['write_id3v1']:
+            v1 = 2
+        else:
+            v1 = 0
+
+        if config.setting['write_id3v23']:
+            tags.update_to_v23()
+            separator = config.setting['id3v23_join_with']
+            tags.save(filename, v2_version=3, v1=v1, v23_sep=separator)
+        else:
+            tags.update_to_v24()
+            tags.save(filename, v2_version=4, v1=v1)
 
     @property
     def new_metadata(self):
@@ -462,7 +538,7 @@ class ID3File(File):
             # If this is a multi-valued field, then it needs to be flattened,
             # unless it's TIPL or TMCL which can still be multi-valued.
 
-            if (len(values) > 1 and not name in ID3File._rtipl_roles
+            if (len(values) > 1 and name not in ID3File._rtipl_roles
                     and not name.startswith("performer:")):
                 values = [join_with.join(values)]
 
@@ -474,10 +550,13 @@ class ID3File(File):
 class MP3File(ID3File):
 
     """MP3 file."""
-    EXTENSIONS = [".mp3", ".mp2"]
+    EXTENSIONS = [".mp3", ".mp2", ".m2a"]
     NAME = "MPEG-1 Audio"
-    _File = mutagen.mp3.MP3
     _IsMP3 = True
+    _File = mutagen.mp3.MP3
+
+    def _get_file(self, filename):
+        return mutagen.mp3.MP3(filename, ID3=compatid3.CompatID3)
 
     def _info(self, metadata, file):
         super(MP3File, self)._info(metadata, file)
@@ -494,6 +573,42 @@ class TrueAudioFile(ID3File):
     NAME = "The True Audio"
     _File = mutagen.trueaudio.TrueAudio
 
+    def _get_file(self, filename):
+        return mutagen.trueaudio.TrueAudio(filename, ID3=compatid3.CompatID3)
+
     def _info(self, metadata, file):
         super(TrueAudioFile, self)._info(metadata, file)
         metadata['~format'] = self.NAME
+
+
+if mutagen.aiff:
+    class AiffFile(ID3File):
+
+        """AIFF file."""
+        EXTENSIONS = [".aiff", ".aif", ".aifc"]
+        NAME = "Audio Interchange File Format (AIFF)"
+        _File = mutagen.aiff.AIFF
+
+        def _get_file(self, filename):
+            return mutagen.aiff.AIFF(filename)
+
+        def _get_tags(self, filename):
+            file = self._get_file(filename)
+            if file.tags is None:
+                file.add_tags()
+            return file.tags
+
+        def _save_tags(self, tags, filename):
+            if config.setting['write_id3v23']:
+                tags.update_to_v23()
+                separator = config.setting['id3v23_join_with']
+                tags.save(filename, v2_version=3, v23_sep=separator)
+            else:
+                tags.update_to_v24()
+                tags.save(filename, v2_version=4)
+
+        def _info(self, metadata, file):
+            super(AiffFile, self)._info(metadata, file)
+            metadata['~format'] = self.NAME
+else:
+    AiffFile = None
